@@ -59,6 +59,7 @@ static void box_test (void);
 static void draw_symbols_test (void);
 static void lr_corner_test (void);
 static void mono_test (void);
+static void wrapped_string_test (void);
 
 typedef struct
 {
@@ -82,6 +83,7 @@ static Menu_Type Root_Menu [] =
      {"Test Mouse", mouse_test},
      {"Box Test", box_test},
      {"Write to Lower Right Corner Test", lr_corner_test},
+     {"Write Wrapped String Test", wrapped_string_test},
      {"Test Low Level Functions", low_level_test},
      {"Test monochrome functions", mono_test},
      {"Quit", quit},
@@ -235,10 +237,13 @@ static void init_colors (void)
    char *fg, *bg;
 
    fg = "black";
-   for (i = 0; i < NUM_COLORS; i++)
+   SLtt_set_color (1, NULL, fg, "default");
+   SLtt_set_color (1+NUM_COLORS, NULL, "default", "fg");
+   for (i = 1; i < NUM_COLORS; i++)
      {
 	bg = Colors[i];
 	SLtt_set_color (i + 1, NULL, fg, bg);
+	SLtt_set_color (i + 1 + NUM_COLORS, NULL, bg, fg);
      }
 }
 
@@ -318,9 +323,9 @@ static void lr_corner_test (void)
    pre_test ("Write to Lower Right Corner Test");
 
    SLsmg_gotorc (SLtt_Screen_Rows-1, SLtt_Screen_Cols-1);
-   SLsmg_write_string ("X");
+   SLsmg_write_string ("Ç");
 
-   write_centered_string ("An 'X' should be in the lower-right corner.",
+   write_centered_string ("A 'Ç' should be in the lower-right corner. (assuming UTF-8)",
 			  SLtt_Screen_Rows/2);
 
    SLsmg_refresh ();
@@ -346,7 +351,7 @@ static void color_test (void)
 	color = color % NUM_COLORS;
 
 	SLsmg_gotorc (row, 0);
-	SLsmg_set_color (0);
+	SLsmg_set_color (color+1+NUM_COLORS);
 	SLsmg_write_string (Colors[color]);
 	color++;
 	SLsmg_set_color (color);
@@ -421,7 +426,7 @@ static void color_test1 (void)
 static void alt_char_test (void)
 {
    int row, col;
-   int ch;
+   SLwchar_Type ch;
    
    pre_test ("Alternate Charset Test");
    
@@ -478,7 +483,7 @@ static Draw_Symbols_Type Draw_Symbols [] =
      {"SLSMG_UARROW_CHAR",		SLSMG_UARROW_CHAR},
      {"SLSMG_BOARD_CHAR",		SLSMG_BOARD_CHAR},
      {"SLSMG_BLOCK_CHAR",		SLSMG_BLOCK_CHAR},
-     {NULL}
+     {NULL, 0}
 };
 
 static void draw_symbols_test (void)
@@ -495,7 +500,7 @@ static void draw_symbols_test (void)
      {
 	SLsmg_gotorc (row, col);
 	SLsmg_set_char_set (1);
-	SLsmg_write_char ((char) d->value);
+	SLsmg_write_char (d->value);
 	SLsmg_set_char_set (0);
 	SLsmg_printf (":%s", d->name);
 	col += 40;
@@ -665,6 +670,60 @@ static void mono_test (void)
    SLtt_Use_Ansi_Colors = c;
 }
 
+static void wrapped_string_test (void)
+{
+   int row, col;
+   unsigned int dr, dc;
+   char *str;
+
+   dc = 15;
+   dr = 10;
+   row = (SLtt_Screen_Rows - (int)dr)/2;
+   col = (SLtt_Screen_Cols - (int)dc)/2;
+
+   pre_test ("Wrapped-string test");
+   
+   str = "This is a string that should be wrapped in a 12x15 cell.\n\
+It even contains a\n\
+couple of newline chacters\n\
+for fun.";
+   
+   SLsmg_fill_region (row-1, col-1, dr+2, dc+2, '.');
+   SLsmg_draw_box (row-1, col-1, dr+2, dc+2);
+   SLsmg_write_wrapped_string ((SLuchar_Type *)str, row, col, dr, dc, 1);
+
+   if (SLutf8_is_utf8_mode ())
+     {
+	unsigned int len;
+	post_test ();
+
+	str = "UTF8: พระปกเกศกองบู๊กู้ขึ้นใหม่\
+สององค์ไซร้โง่เขลาเบาปัญญา\
+บ้านเมืองจึงวิปริตเป็นนักหนา\
+หมายจะฆ่ามดชั่วตัวสำคัญ\
+รับหมาป่าเข้ามาเลยอาสัญ\
+ใช้สาวนั้นเป็นชนวนชื่นชวนใจ\
+ช่างอาเพศจริงหนาฟ้าร้องไห้\
+ฤๅหาใครค้ำชูกู้บรรลังก์ ฯ";
+
+	SLsmg_fill_region (row-1, col-1, dr+2, dc+2, '.');
+	SLsmg_draw_box (row-1, col-1, dr+2, dc+2);
+	SLsmg_write_wrapped_string ((SLuchar_Type *)str, row, col, dr, dc, 1);
+	
+	post_test ();
+
+	col = 5;
+	str = "รั\tบ\tหมาป่าเข้ามาเลยอาสัญ\xAB";
+	SLsmg_cls ();
+	SLsmg_gotorc (row, col);
+	len = SLsmg_strwidth ((SLuchar_Type*)str, (SLuchar_Type*)str + strlen (str));
+	SLsmg_write_string (str);
+	SLsmg_gotorc (row+1, col+len);
+	SLsmg_write_string ("^---End of above line should be here");
+     }
+   post_test ();
+}
+
 static void low_level_test (void)
 {
    int mid, bot;
