@@ -1,3 +1,24 @@
+/*
+Copyright (C) 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.
+*/
+
 #include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
@@ -11,21 +32,33 @@
 # include <unistd.h>
 #endif
 #include <string.h>
+#include <signal.h>
 #include <slang.h>
 
-static char *Slsh_Version = "0.5-0";
+static char *Slsh_Version = "0.7.7-1";
 #define SLSHRC_FILE "slsh.rc"
+#include "slsh.h"
 
-#ifdef REAL_UNIX_SYSTEM
-# define DEFAULT_LIBRARY_PATH "/usr/local/etc:/usr/local/slsh:/etc:/etc/slsh";
-# define USER_SLSHRC ".slshrc"
-#else 
-# define DEFAULT_LIBRARY_PATH NULL
-# define USER_SLSHRC "slsh.rc"
+#ifndef SLSH_CONF_DIR_ENV
+# define SLSH_CONF_DIR_ENV "SLSH_CONF_DIR"
 #endif
 
-static char *Slsh_Library_Path;
-static int Verbose_Loading;
+#ifndef SLSH_PATH_ENV
+# define SLSH_PATH_ENV "SLSH_CONF_DIR"
+#endif
+
+#ifndef SLSH_LIB_DIR_ENV
+# define SLSH_LIB_DIR_ENV "SLSH_LIB_DIR"
+#endif
+
+#if defined(REAL_UNIX_SYSTEM) || defined(__APPLE__)
+/* # define DEFAULT_LIBRARY_PATH "/usr/local/share/slsh:/usr/local/lib/slsh:/usr/share/slsh:/usr/lib/slsh"; */
+# define DEFAULT_CONF_PATH "/usr/local/etc:/usr/local/etc/slsh:/etc:/etc/slsh";
+# define USER_SLSHRC ".slshrc"
+#else 
+# define DEFAULT_CONF_PATH NULL
+# define USER_SLSHRC "slsh.rc"
+#endif
 
 #ifdef __os2__
 # ifdef __IBMC__
@@ -159,38 +192,17 @@ static void c_exit (int *code)
    while (AtExit_Hooks != NULL)
      {
 	AtExit_Type *next = AtExit_Hooks->next;
-	if (SLang_Error == 0)
+	if (SLang_get_error () == 0)
 	  (void) SLexecute_function (AtExit_Hooks->nt);
 
 	SLfree ((char *) AtExit_Hooks);
 	AtExit_Hooks = next;
      }
+   if (SLang_get_error ())
+     SLang_restart (1);
    exit (*code);
 }
 
-static int set_lib_path (char *path)
-{
-   if (path == NULL)
-     return 0;
-
-   path = SLang_create_slstring (path);
-   if (path == NULL)
-     return -1;
-   
-   if (Slsh_Library_Path != NULL)
-     SLang_free_slstring (Slsh_Library_Path);
-   
-   Slsh_Library_Path = path;
-   return 0;
-}
-
-static char *get_lib_path (void)
-{
-   if (Slsh_Library_Path == NULL)
-     return "";
-   return Slsh_Library_Path;
-}
-
 
 static void stat_mode_to_string (void)
 {
@@ -245,81 +257,110 @@ static void stat_mode_to_string (void)
    (void) SLang_push_string (mode_string);
 }
 
+
+#ifdef __WIN32__
+
+/* Root/ 
+ * 	bin/
+ * 		slsh.exe 
+ * 	slsh/ 
+ * 		slsh.rc 
+ * 		autoload.sl 
+ * 		( ... )
+ */
+static char *get_win32_root (void)
+{
+   static char base_path[MAX_PATH] = "";
+
+   if (*base_path == 0)
+     {
+	if (GetModuleFileName(NULL, base_path, MAX_PATH-10) > 0) 
+	  {
+	     /* drop file name */
+	     char *p = strrchr(base_path, '\\');
+	     if (p != NULL)
+	       {
+		  *p = 0;
+		  /* drop also 'bin' */
+		  p = strrchr(base_path, '\\');
+		  if (p != NULL)
+		    {
+		       strcpy(p, "\\slsh\\");
+		    }
+	       }
+	  }
+     }
+   
+   return base_path;
+}
+#endif
+
+static int Verbose_Loading;
+
 static int try_to_load_file (char *path, char *file, char *ns)
 {
+   int status;
+
    if (path == NULL)
      path = ".";
 
-   file = SLpath_find_file_in_path (path, file);
-   if (file == NULL)
-     return 0;
-   
-   if (Verbose_Loading)
-     fprintf (stdout, "loading %s\n", file);
-
-   if (0 == SLns_load_file (file, ns))
+   if (file != NULL)
      {
-	SLfree (file);
-	return 1;
+	file = SLpath_find_file_in_path (path, file);
+	if (file == NULL)
+	  return 0;
      }
+   /* otherwise use stdin */
+
+   status = SLns_load_file (file, ns);
    SLfree (file);
+   if (status == 0)
+     return 1;
    return -1;
 }
 
 
-static int load_file (void)
-{
-   char *file;
-   char *ns = NULL;
-   int status;
-
-   if (SLang_Num_Function_Args == 2)
-     {
-	if (-1 == SLang_pop_slstring (&ns))
-	  return -1;
-     }
-
-   if (-1 == SLang_pop_slstring (&file))
-     {
-	SLang_free_slstring (ns);
-	return -1;
-     }
-   
-   status = try_to_load_file (Slsh_Library_Path, file, ns);
-   if (status != 1)
-     SLang_verror (SL_OBJ_NOPEN, "Unable to open %s", file);
-   SLang_free_slstring (file);
-   SLang_free_slstring (ns);
-   return status;
-}
-
-static int load_startup_file (void)
+static int load_startup_file (int is_interactive)
 {
    char *dir;
+   int status;
 
-   dir = getenv ("SLSH_CONF_DIR");
+   dir = getenv (SLSH_CONF_DIR_ENV);
    if (dir == NULL)
-     dir = getenv ("SLSH_LIB_DIR");
+     dir = getenv (SLSH_LIB_DIR_ENV);
 
    if (NULL == dir)
      {
-#ifdef SLSH_LIB_DIR
-	dir = SLSH_LIB_DIR;
+#ifdef SLSH_CONF_DIR
+	dir = SLSH_CONF_DIR;
 	if (dir != NULL)
 	  {
-	     int status = try_to_load_file (dir, SLSHRC_FILE, NULL);
+	     status = try_to_load_file (dir, SLSHRC_FILE, NULL);
 	     if (status == -1)
 	       return -1;
 	     if (status == 1)
 	       return 0;
 	  }
 #endif
-	dir = DEFAULT_LIBRARY_PATH;
+
+#ifdef __WIN32__
+	if (NULL == (dir = get_win32_root ()))
+	  dir = DEFAULT_CONF_PATH;
+#else	
+	dir = DEFAULT_CONF_PATH;
+#endif
      }
 
-   if (-1 == try_to_load_file (dir, SLSHRC_FILE, NULL))
+   if (-1 == (status = try_to_load_file (dir, SLSHRC_FILE, NULL)))
      return -1;
 
+   if ((status == 0) 
+       && (Verbose_Loading || is_interactive))
+     {
+	SLang_vmessage ("*** Installation Problem?  Unable to find the %s config file.",
+			SLSHRC_FILE);
+     }
+     
    return 0;
 }
 
@@ -342,42 +383,89 @@ static int is_script (char *file)
 }
 #endif
 
+static int setup_paths (void)
+{
+   char *libpath;
+
+   if (NULL == (libpath = getenv (SLSH_PATH_ENV)))
+     {
+#ifdef SLSH_PATH
+	libpath = SLSH_PATH;
+#else
+# ifdef __WIN32__
+	libpath = get_win32_root ();
+# endif
+#endif
+     }
+
+   return SLpath_set_load_path (libpath);
+}
+
 /* Create the Table that S-Lang requires */
 static SLang_Intrin_Fun_Type Intrinsics [] =
 {
-   MAKE_INTRINSIC_0("evalfile", load_file, INT_TYPE),
-   MAKE_INTRINSIC_S("set_lib_path", set_lib_path, VOID_TYPE),
-   MAKE_INTRINSIC_0("get_lib_path", get_lib_path, SLANG_STRING_TYPE),
    MAKE_INTRINSIC_I("exit", c_exit, VOID_TYPE),
    MAKE_INTRINSIC_1("atexit", at_exit, VOID_TYPE, SLANG_REF_TYPE),
+   MAKE_INTRINSIC_S("slsh_readline", slsh_readline_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_S("slsh_readline_noecho", slsh_readline_noecho_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_0("slsh_set_prompt_hook", slsh_set_prompt_hook, VOID_TYPE),
+   MAKE_INTRINSIC_0("slsh_get_prompt_hook", slsh_get_prompt_hook, VOID_TYPE),
    MAKE_INTRINSIC_0("stat_mode_to_string", stat_mode_to_string, VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
 
 static void usage (void)
 {
+   char *libpath;
    fprintf (stderr, "\
 Usage: slsh [OPTIONS] [-|file [args...]]\n\
  --help           Print this help\n\
  --version        Show slsh version information\n\
  -g               Compile with debugging code, tracebacks, etc\n\
  -n               Don't load personal init file\n\
- -i init-file     Use this file instead of default\n\
+ --init file      Use this file instead of ~/%s\n",
+	    USER_SLSHRC);
+   fprintf (stderr, "\
+ --no-readline    Do not use readline\n\
+ -i               Force interactive input\n\
+ -t               Test mode.  If slsh_main exits, do not call it\n\
  -v               Show verbose loading messages\n\
+\n\
+  Note: - and -i are mutually exclusive\n\
+\n\
 "
 	    );
+   libpath = SLpath_get_load_path ();
+   fprintf (stderr, "Default search path: %s\n", (libpath == NULL) ? "" : libpath);
+   SLang_free_slstring (libpath);
+   
    exit (1);
 }
 
-static void version (void)
+static void output_version (void)
 {
-   fprintf (stdout, "slsh version %s\n", Slsh_Version);
-   fprintf (stdout, "S-Lang Library Version: %s\n", SLang_Version_String);
+   fprintf (stdout, "slsh version %s; ", Slsh_Version);
+   fprintf (stdout, "S-Lang version: %s\n", SLang_Version_String);
    if (SLANG_VERSION != SLang_Version)
      {
 	fprintf (stdout, "\t** Note: This program was compiled against version %s.\n",
 		 SLANG_VERSION_STRING);
      }
+}
+
+static int output_copyright (void)
+{
+   output_version ();
+   fprintf (stdout, "Copyright (C) 2005-2006 John E. Davis <jed@jedsoft.org>\r\n");
+   fprintf (stdout, "This is free software with ABSOLUTELY NO WARRANTY.\r\n");
+   fprintf (stdout, "\n");
+   
+   return 0;
+}
+
+static void version (void)
+{
+   output_version ();
    exit (0);
 }
 
@@ -385,6 +473,13 @@ int main (int argc, char **argv)
 {
    char *file = NULL;
    char *init_file = USER_SLSHRC;
+   char *init_file_dir;
+   int exit_val;
+   int is_interactive = 0;
+   int use_readline = 1;
+   int test_mode = 0;
+
+   (void) SLutf8_enable (-1);
 
    if ((-1 == SLang_init_all ())
        || (-1 == SLang_init_array_extra ())
@@ -395,6 +490,34 @@ int main (int argc, char **argv)
 	return 1;
      }
 
+#ifdef SIGPIPE
+   (void) SLsignal (SIGPIPE, SIG_IGN);
+#endif
+
+   /* FIXME for other systems */
+#ifdef __WIN32__
+   init_file_dir = getenv ("USERPROFILE");
+#else
+   init_file_dir = getenv ("HOME");
+#endif
+
+   if (-1 == setup_paths ())
+     return -1;
+
+#ifdef __WIN32__
+   /* hack a directory for modules */
+     {
+	char *slshdir;
+	char buffer[MAX_PATH] = "";
+	
+	slshdir = get_win32_root();
+	if (slshdir != NULL)
+	  strcpy(buffer, slshdir);
+	strcat(buffer, "modules\\");
+	SLang_set_module_load_path (buffer);
+     }
+#endif
+
    while (argc > 1)
      {
 	if (0 == strcmp (argv[1], "--version"))
@@ -403,9 +526,18 @@ int main (int argc, char **argv)
 	if (0 == strcmp (argv[1], "--help"))
 	  usage ();
 
+	if (0 == strcmp (argv[1], "-i"))
+	  {
+	     argc--;
+	     argv++;
+	     is_interactive = 1;
+	     continue;
+	  }
+
 	if (0 == strcmp (argv[1], "-g"))
 	  {
 	     SLang_generate_debug_info (1);
+	     SLang_Traceback = 1;
 	     argc--;
 	     argv++;
 	     continue;
@@ -418,19 +550,37 @@ int main (int argc, char **argv)
 	     argv++;
 	     continue;
 	  }
+
+	if (0 == strcmp (argv[1], "-t"))
+	  {
+	     test_mode = 1;
+	     argc--;
+	     argv++;
+	     continue;
+	  }
 	
 	if (0 == strcmp (argv[1], "-v"))
 	  {
+	     (void) SLang_load_file_verbose (1);
 	     Verbose_Loading = 1;
 	     argc--;
 	     argv++;
 	     continue;
 	  }
 	
-	if ((0 == strcmp (argv[1], "-i"))
+	if (0 == strcmp (argv[1], "--no-readline"))
+	  {
+	     use_readline = 0;
+	     argc--;
+	     argv++;
+	     continue;
+	  }
+
+	if ((0 == strcmp (argv[1], "--init"))
 	    && (argc > 2))
 	  {
 	     init_file = argv[2];
+	     init_file_dir = NULL;
 	     argc -= 2;
 	     argv += 2;
 	     continue;
@@ -440,46 +590,66 @@ int main (int argc, char **argv)
 
    if (argc == 1)
      {
-	if (0 == isatty (fileno(stdin)))
-	  file = NULL;
-	else
-	  usage ();
+	is_interactive = (isatty (fileno(stdin)) && isatty (fileno(stdout)));
+	file = NULL;
      }
    else
      {
 	file = argv[1];
 	if (0 == strcmp (file, "-"))
-	  file = NULL;
-#if 0
-	if (is_script (file))
 	  {
-	     argv++;
-	     argc--;
+	     if (is_interactive)
+	       usage ();
+	     file = NULL;
 	  }
-#else
+	
 	argc--;
 	argv++;
-#endif
      }
+
+   if ((is_interactive == 0)
+       && (SLang_Version < SLANG_VERSION))
+     {
+	fprintf (stderr, "***Warning: Executable compiled against S-Lang %s but linked to %s\n",
+		 SLANG_VERSION_STRING, SLang_Version_String);
+	fflush (stderr);
+     }
+
+   (void) slsh_use_readline (use_readline);
+
    /* fprintf (stdout, "slsh: argv[0]=%s\n", argv[0]); */
    if (-1 == SLang_set_argc_argv (argc, argv))
      return 1;
 
-   if (-1 == set_lib_path (getenv ("SLSH_PATH")))
-     return 1;
+   if (is_interactive)
+     (void) SLdefine_for_ifdef ("__INTERACTIVE__");
 
-   /* Turn on traceback generation */
-   SLang_Traceback = 1;
+   if (-1 == load_startup_file (is_interactive))
+     return SLang_get_error ();
 
-   if (-1 == load_startup_file ())
-     return SLang_Error;
-   
    if ((init_file != NULL)
-       && (-1 == try_to_load_file (getenv ("HOME"), USER_SLSHRC, NULL)))
-     return SLang_Error;
+       && (-1 == try_to_load_file (init_file_dir, init_file, NULL)))
+     return SLang_get_error ();
 
    /* Now load an initialization file and exit */
-   (void) SLang_load_file (file);
-   (void) SLang_run_hooks ("slsh_main", 0);
-   return SLang_Error;
+   if ((file != NULL) || (is_interactive == 0))
+     {
+	if (0 == try_to_load_file (NULL, file, NULL))
+	  {
+	     fprintf (stderr, "%s: file not found\n", file);
+	     exit (1);
+	  }
+	if (test_mode == 0)
+	  (void) SLang_run_hooks ("slsh_main", 0);
+     }
+
+   if (is_interactive)
+     {
+	output_copyright ();
+	(void) slsh_interactive ();
+     }
+
+   exit_val = SLang_get_error ();
+   c_exit (&exit_val);
+   return SLang_get_error ();
 }
