@@ -174,7 +174,9 @@ typedef enum
    /* SLANG_BC_ARG_LVARIABLE	= 0x72, */
    SLANG_BC_UNUSED_0x72		= 0x72,
    SLANG_BC_EARG_LVARIABLE	= 0x73,
-   SLANG_BC_LINE_NUM		= 0x74,
+#define USE_BC_LINE_NUM	0	       /* used in slang.c */
+   /* SLANG_BC_LINE_NUM		= 0x74, */
+   SLANG_BC_UNUSED_0x74		= 0x74,
    SLANG_BC_BOS			= 0x75,
    SLANG_BC_EOS			= 0x76,
    SLANG_BC_UNUSED_0x77		= 0x77,
@@ -671,12 +673,15 @@ extern char *_pSLang_current_function_name (void);
 extern int _pSLang_trace_fun(char *);
 
 /* This is a bitmapped variable */
-extern int _pSLang_Compile_Line_Num_Info;
+/* extern int _pSLang_Compile_Line_Num_Info; */
 #if SLANG_HAS_BOSEOS
 extern int _pSLang_Compile_BOSEOS;
+extern int _pSLang_Compile_BOFEOF;
 extern int _pSLang_init_boseos (void);
 extern int _pSLcall_bos_handler (char *, int);
 extern int _pSLcall_eos_handler (void);
+extern int _pSLcall_bof_handler (char *, char *);
+extern int _pSLcall_eof_handler (void);
 extern int _pSLcall_debug_hook (char *file, int linenum);
 /* extern int _pSLcall_debug_hook (char *file, int linenum, char *funct); */
 #endif
@@ -1288,6 +1293,8 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define _INLINE_WILDCARD_ARRAY_TOKEN	0xE3
 #define _INLINE_LIST_TOKEN		0xE4
 
+#define ESC_STRING_DOLLAR_TOKEN		0xF0
+
 #define BOS_TOKEN			0xFA
 #define EOS_TOKEN			0xFB
 #define LINE_NUM_TOKEN			0xFC
