@@ -3,19 +3,37 @@
  * That is, a keymap is simple a mapping of strings (keys from tty) to
  * structures.  Also included are routines for managing the keymaps.
  */
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
 #include "slang.h"
 #include "_slang.h"
 
-/* We need a define a rule for upperand lower case chars that user cannot
+/* FIXME: I need to add a copy-method so that copy-keymap will work with
+ * any object bound to the keysequence.
+ */
+
+/* We need a define a rule for upper and lower case chars that user cannot
    change!  This could be a problem for international chars! */
 
 #define UPPER_CASE_KEY(x) (((x) >= 'a') && ((x) <= 'z') ? (x) - 32 : (x))
@@ -26,7 +44,89 @@ int SLang_Key_TimeOut_Flag = 0;	       /* true if more than 1 sec has elapsed
 
 int SLang_Last_Key_Char;
 
-SLKeyMap_List_Type SLKeyMap_List[SLANG_MAX_KEYMAPS];
+SLkeymap_Type *SLKeyMap_List_Root;
+
+typedef struct _Key_Methods_Type
+{
+   int type;
+   void (*free_method) (int, VOID_STAR);
+}
+Key_Methods_Type;
+
+static unsigned int Num_Key_Methods;
+
+static void free_interpret_key (int type, VOID_STAR addr)
+{
+   (void) type;
+   
+   SLang_free_slstring (*(char **) addr);
+}
+
+#define SLKEYMAP_MAX_TYPES 16
+static Key_Methods_Type Key_Methods_Table[SLKEYMAP_MAX_TYPES] = 
+{
+   {SLKEY_F_INTERPRET, free_interpret_key}
+};
+
+static Key_Methods_Type *find_key_methods (int type)
+{
+   Key_Methods_Type *k, *kmax;
+   
+   k = Key_Methods_Table;
+   kmax = k + Num_Key_Methods;
+
+   while (k < kmax)
+     {
+	if (k->type == type)
+	  return k;
+	
+	k++;
+     }
+   return NULL;
+}
+
+static Key_Methods_Type *alloc_key_methods (int type)
+{
+   Key_Methods_Type *k;
+
+   if (Num_Key_Methods >= SLKEYMAP_MAX_TYPES)
+     {
+	SLang_verror (SL_LimitExceeded_Error, "Maximum number of keymap types exceeded");
+	return NULL;
+     }
+   k = Key_Methods_Table + Num_Key_Methods;
+   k->type = type;
+   Num_Key_Methods++;
+   return k;
+}
+
+int SLkm_set_free_method (int type, void (*f)(int, VOID_STAR))
+{
+   Key_Methods_Type *k;
+
+   if ((NULL == (k = find_key_methods (type)))
+       && (NULL == (k = alloc_key_methods (type))))
+     return -1;
+   
+   k->free_method = f;
+   return 0;
+}
+
+/* This frees the function associated with the key, but does not does not
+ * do anything with the keysequence (str) field.
+ */
+static void free_key_function (SLang_Key_Type *key)
+{
+   Key_Methods_Type *k;
+   
+   k = find_key_methods (key->type);
+   if ((k != NULL) && (k->free_method != NULL))
+     (*k->free_method) (key->type, &key->f);
+   
+   key->f.f = NULL;
+   key->type = 0;
+   /* *key->str = 0;  Let caller do this */
+}
 
 static SLang_Key_Type *malloc_key(unsigned char *str)
 {
@@ -40,28 +140,25 @@ static SLang_Key_Type *malloc_key(unsigned char *str)
    return(neew);
 }
 
-static SLKeyMap_List_Type *add_keymap (char *name, SLang_Key_Type *map)
+static SLkeymap_Type *add_keymap (char *name, SLang_Key_Type *map)
 {
-   int i;
+   SLkeymap_Type *km;
 
-   for (i = 0; i < SLANG_MAX_KEYMAPS; i++)
+   if (NULL == (km = (SLkeymap_Type *)SLcalloc (1, sizeof(SLkeymap_Type))))
+     return NULL;
+
+   if (NULL == (km->name = SLang_create_slstring (name)))
      {
-	if (SLKeyMap_List[i].keymap == NULL)
-	  {
-	     if (NULL == (name = SLang_create_slstring (name)))
-	       return NULL;
-
-	     SLKeyMap_List[i].keymap = map;
-	     SLKeyMap_List[i].name = name;
-	     return &SLKeyMap_List[i];
-	  }
+	SLfree ((char *)km);
+	return NULL;
      }
-   SLang_Error = SL_UNKNOWN_ERROR;
-   /* SLang_doerror ("Keymap quota exceeded."); */
-   return NULL;
+   km->keymap = map;
+   km->next = SLKeyMap_List_Root;
+   SLKeyMap_List_Root = km;
+   return km;
 }
 
-FVOID_STAR SLang_find_key_function(char *name, SLKeyMap_List_Type *keymap)
+FVOID_STAR SLang_find_key_function(char *name, SLkeymap_Type *keymap)
 {
    SLKeymap_Function_Type *fp = keymap -> functions;
    char ch = *name;
@@ -190,10 +287,10 @@ static int key_string_compare (unsigned char *a, unsigned char *b, unsigned int
    return 0;
 }
 
-static char *Define_Key_Error = "Inconsistency in define key.";
+#define KEY_DEFINITION_ERROR_MSG "Inconsistent key-definition"
 
 /* This function also performs an insertion in an ordered way. */
-static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp)
+static int find_the_key (char *s, SLkeymap_Type *kml, SLang_Key_Type **keyp)
 {
    unsigned char ch;
    unsigned int str_len;
@@ -215,12 +312,11 @@ static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp
      {
 	if (key->next != NULL)
 	  {
-	     SLang_doerror (Define_Key_Error);
+	     SLang_verror (SL_INVALID_PARM, KEY_DEFINITION_ERROR_MSG);
 	     return -2;
 	  }
 
-	if (key->type == SLKEY_F_INTERPRET)
-	  SLang_free_slstring (key->f.s);
+	free_key_function (key);
 
 	key->str[0] = str_len;
 	key->str[1] = ch;
@@ -252,12 +348,11 @@ static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp
 	       {
 		  if (key_len != str_len)
 		    {
-		       SLang_doerror (Define_Key_Error);
+		       SLang_verror (SL_INVALID_PARM, KEY_DEFINITION_ERROR_MSG);
 		       return -2;
 		    }
-
-		  if (key->type == SLKEY_F_INTERPRET)
-		    SLang_free_slstring (key->f.s);
+		  
+		  free_key_function (key);
 
 		  *keyp = key;
 		  return 0;
@@ -276,7 +371,7 @@ static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp
 }
 
 /* returns -2 if inconsistent, -1 if malloc error, 0 upon success */
-int SLkm_define_key (char *s, FVOID_STAR f, SLKeyMap_List_Type *kml)
+int SLkm_define_key (char *s, FVOID_STAR f, SLkeymap_Type *kml)
 {
    SLang_Key_Type *key;
    unsigned int type = SLKEY_F_INTRINSIC;
@@ -291,7 +386,7 @@ int SLkm_define_key (char *s, FVOID_STAR f, SLKeyMap_List_Type *kml)
    return 0;
 }
 
-int SLang_define_key (char *s, char *funct, SLKeyMap_List_Type *kml)
+int SLang_define_key (char *s, char *funct, SLkeymap_Type *kml)
 {
    SLang_Key_Type *key;
    FVOID_STAR f;
@@ -318,7 +413,7 @@ int SLang_define_key (char *s, char *funct, SLKeyMap_List_Type *kml)
    return 0;
 }
 
-int SLkm_define_keysym (char *s, unsigned int keysym, SLKeyMap_List_Type *kml)
+int SLkm_define_keysym (char *s, unsigned int keysym, SLkeymap_Type *kml)
 {
    SLang_Key_Type *key;
    int ret;
@@ -333,7 +428,21 @@ int SLkm_define_keysym (char *s, unsigned int keysym, SLKeyMap_List_Type *kml)
    return 0;
 }
 
-SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
+int SLkm_define_slkey (char *s, SLang_Name_Type *nt, SLkeymap_Type *kml)
+{
+   SLang_Key_Type *key;
+   int ret;
+
+   ret = find_the_key (s, kml, &key);
+   if ((ret != 0) || (key == NULL))
+     return ret;
+
+   key->type = SLKEY_F_SLANG;
+   key->f.slang_fun = nt;
+   return 0;
+}
+
+SLang_Key_Type *SLang_do_key(SLkeymap_Type *kml, int (*getkey)(void))
 {
    register SLang_Key_Type *key, *next, *kmax;
    unsigned int len;
@@ -452,7 +561,7 @@ SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
    return NULL;
 }
 
-void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
+void SLang_undefine_key(char *s, SLkeymap_Type *kml)
 {
    int n, i;
    SLang_Key_Type *key, *next, *last, *key_root, *keymap;
@@ -473,8 +582,7 @@ void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
 	next = key->next;
 	if (0 == SLMEMCMP ((char *)(key->str + 1), (char *) str, n))
 	  {
-	     if (key->type == SLKEY_F_INTERPRET)
-	       SLang_free_slstring (key->f.s);
+	     free_key_function (key);
 
 	     SLfree((char *) key);
 	     last->next = next;
@@ -485,9 +593,8 @@ void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
 
    if (n == 1)
      {
-	*key_root->str = 0;
-	key_root->f.f = NULL;
-	key_root->type = 0;
+	free_key_function (key_root);
+	key_root->str[0] = 0;
      }
 }
 
@@ -501,7 +608,7 @@ char *SLang_make_keystring(unsigned char *s)
 
    if (n > SLANG_MAX_KEYMAP_KEY_SEQ)
      {
-	SLang_verror (SL_INVALID_PARM, "Key sequence is too long");
+	SLang_verror (SL_BUILTIN_LIMIT_EXCEEDED, "Key sequence is too long");
 	return NULL;
      }
 
@@ -520,7 +627,7 @@ char *SLang_make_keystring(unsigned char *s)
    return(buf);
 }
 
-static SLang_Key_Type *copy_keymap(SLKeyMap_List_Type *kml)
+static SLang_Key_Type *copy_keymap(SLkeymap_Type *kml)
 {
    int i;
    SLang_Key_Type *neew, *old, *new_root, *km;
@@ -536,6 +643,7 @@ static SLang_Key_Type *copy_keymap(SLKeyMap_List_Type *kml)
 	old = &(km[i]);
 	neew = &(new_root[i]);
 
+	/* FIXME: A copy-method needs to be defined here and below */
 	if (old->type == SLKEY_F_INTERPRET)
 	  neew->f.s = SLang_create_slstring (old->f.s);
 	else
@@ -563,10 +671,10 @@ static SLang_Key_Type *copy_keymap(SLKeyMap_List_Type *kml)
    return(new_root);
 }
 
-SLKeyMap_List_Type *SLang_create_keymap(char *name, SLKeyMap_List_Type *map)
+SLkeymap_Type *SLang_create_keymap(char *name, SLkeymap_Type *map)
 {
    SLang_Key_Type *neew;
-   SLKeyMap_List_Type *new_map;
+   SLkeymap_Type *new_map;
 
    if ((NULL == (neew = copy_keymap(map)))
        || (NULL == (new_map = add_keymap(name, neew)))) return NULL;
@@ -576,20 +684,19 @@ SLKeyMap_List_Type *SLang_create_keymap(char *name, SLKeyMap_List_Type *map)
    return new_map;
 }
 
-SLKeyMap_List_Type *SLang_find_keymap(char *name)
+SLkeymap_Type *SLang_find_keymap(char *name)
 {
-   SLKeyMap_List_Type *kmap, *kmap_max;
+   SLkeymap_Type *kmap;
 
-   kmap = SLKeyMap_List;
-   kmap_max = kmap + SLANG_MAX_KEYMAPS;
+   kmap = SLKeyMap_List_Root;
 
-   while (kmap < kmap_max)
+   while (kmap != NULL)
      {
 	if ((kmap->name != NULL)
 	    && (0 == strcmp (kmap->name, name)))
 	  return kmap;
 
-	kmap++;
+	kmap = kmap->next;
      }
    return NULL;
 }
