@@ -1,9 +1,23 @@
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -125,7 +139,7 @@ static void thread_getkey ()
       KbdCharIn(&keyInfo, IO_NOWAIT, 0);       /* get a character	*/
       if (keyInfo.fbStatus & 0x040) {          /* found a char process it */
 	if (keyInfo.chChar == SLang_Abort_Char) {
-	  if (SLang_Ignore_User_Abort == 0) SLang_Error = SL_USER_BREAK;
+	  if (SLang_Ignore_User_Abort == 0) SLang_set_error (SL_USER_BREAK);
 	  SLKeyBoard_Quit = 1;
 	}
 	n = (endBuf + 1) % BUFFER_LEN;
@@ -163,8 +177,9 @@ int SLang_init_tty(int abort_char, int dum2, int dum3)
 
   (void) dum2; (void) dum3;
    if (abort_char == -1) abort_char = 3;   /* ^C */
-  SLang_Abort_Char = abort_char;
-
+   SLang_Abort_Char = abort_char;
+   SLKeyBoard_Quit = 0;
+   
   /*  set ^C off */
   signal (SIGINT, SIG_IGN);
   signal (SIGBREAK, SIG_IGN);
@@ -218,7 +233,7 @@ void SLang_reset_tty (void)
 
 /* sleep for *tsecs tenths of a sec waiting for input */
 
-int _SLsys_input_pending(int tsecs)
+int _pSLsys_input_pending(int tsecs)
 {
    if (keyWaiting()) return 1;
 
@@ -238,7 +253,7 @@ int _SLsys_input_pending(int tsecs)
    return (tsecs);
 }
 
-unsigned int _SLsys_getkey ()
+unsigned int _pSLsys_getkey ()
 {
    unsigned int c;
    unsigned char scan;
@@ -246,7 +261,7 @@ unsigned int _SLsys_getkey ()
    int tsecs = 300;
 
    if (!keyWaiting())
-     while (!_SLsys_input_pending(tsecs));
+     while (!_pSLsys_input_pending(tsecs));
 
    /* read codes from buffer */
    RequestSem();
@@ -263,7 +278,7 @@ unsigned int _SLsys_getkey ()
 
       case 0xE0:
       case 0:
-	c = _SLpc_convert_scancode (scan, 0, 1);
+	c = _pSLpc_convert_scancode (scan, 0, 1);
 	break;
 
       default:
