@@ -1,10 +1,25 @@
 /* Pack objects as a binary string */
-/* Copyright (c) 1998, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 #include <ctype.h>
@@ -21,8 +36,9 @@
 
 /* format description:
  *
- *    s = string (null padded)
+ *    s = string (null padded, except when unpacking)
  *    S = string (space padded)
+ *    z = string (null padded)
  *    c = signed char
  *    C = unsigned char
  *    h = short
@@ -35,6 +51,8 @@
  *    J = 16 bit unsigned integer (short)
  *    k = 32 bit signed integer (long)
  *    K = 32 bit unsigned integer (long)
+ *    q = 64 bit unsigned integer (long)
+ *    Q = 64 bit unsigned integer (long)
  *    f = float (native format)
  *    F = 32 bit double
  *    d = double (native format)
@@ -53,7 +71,7 @@ static int Native_Byte_Order = NATIVE_ORDER;
 typedef struct
 {
    char format_type;
-   unsigned char data_type;
+   SLtype data_type;
    unsigned int repeat;
    unsigned int sizeof_type;
    char pad;
@@ -62,7 +80,7 @@ typedef struct
 }
 Format_Type;
 
-static int get_int_type_for_size (unsigned int size, unsigned char *s, unsigned char *u)
+static int get_int_type_for_size (unsigned int size, SLtype *s, SLtype *u)
 {
    if (sizeof (int) == size)
      {
@@ -84,7 +102,14 @@ static int get_int_type_for_size (unsigned int size, unsigned char *s, unsigned
 	if (u != NULL) *u = SLANG_ULONG_TYPE;
 	return 1;
      }
-
+#ifdef HAVE_LONG_LONG
+   if (sizeof (long long) == size)
+     {
+	if (s != NULL) *s = SLANG_LLONG_TYPE;
+	if (u != NULL) *u = SLANG_ULLONG_TYPE;
+	return 1;
+     }
+#endif
    if (s != NULL) *s = 0;
    if (u != NULL) *u = 0;
    SLang_verror (SL_NOT_IMPLEMENTED,
@@ -92,7 +117,7 @@ static int get_int_type_for_size (unsigned int size, unsigned char *s, unsigned
    return -1;
 }
 
-static int get_float_type_for_size (unsigned int size, unsigned char *s)
+static int get_float_type_for_size (unsigned int size, SLtype *s)
 {
    if (sizeof (float) == size)
      {
@@ -115,7 +140,7 @@ static int parse_a_format (char **format, Format_Type *ft)
 {
    char *f;
    char ch;
-   unsigned repeat;
+   unsigned int repeat;
 
    f = *format;
 
@@ -167,7 +192,7 @@ static int parse_a_format (char **format, Format_Type *ft)
 	     /* Check overflow */
 	     if (repeat != repeat10 / 10)
 	       {
-		  SLang_verror (SL_OVERFLOW,
+		  SLang_verror (SL_ARITH_OVERFLOW_ERROR,
 				"Repeat count too large in [un]pack format");
 		  return -1;
 	       }
@@ -234,6 +259,16 @@ static int parse_a_format (char **format, Format_Type *ft)
 	ft->data_type = SLANG_ULONG_TYPE;
 	ft->sizeof_type = sizeof (unsigned long);
 	break;
+#ifdef HAVE_LONG_LONG
+      case 'm':
+	ft->data_type = SLANG_LLONG_TYPE;
+	ft->sizeof_type = sizeof (long long);
+	break;
+      case 'M':
+	ft->data_type = SLANG_ULLONG_TYPE;
+	ft->sizeof_type = sizeof (unsigned long long);
+	break;
+#endif
 
 	/* 16 bit ints */
       case 'j':
@@ -258,6 +293,18 @@ static int parse_a_format (char **format, Format_Type *ft)
 	if (-1 == get_int_type_for_size (4, NULL, &ft->data_type))
 	  return -1;
 	break;
+	
+	/* 64 bit ints */
+      case 'q':
+	ft->sizeof_type = 8;
+	if (-1 == get_int_type_for_size (4, &ft->data_type, NULL))
+	  return -1;
+	break;
+      case 'Q':
+	ft->sizeof_type = 8;
+	if (-1 == get_int_type_for_size (4, NULL, &ft->data_type))
+	  return -1;
+	break;
 
       case 'x':
 	ft->sizeof_type = 1;
@@ -279,6 +326,7 @@ static int parse_a_format (char **format, Format_Type *ft)
 	ft->pad = ' ';
       case 'a':
       case 's':
+      case 'z':
 	ft->data_type = SLANG_BSTRING_TYPE;
 	ft->sizeof_type = 1;
 	ft->is_scalar = 0;
@@ -485,7 +533,11 @@ pack_according_to_format (char *format, unsigned int nitems)
 	memcpy ((char *)b, (char *)ptr, num);
 	b += num;
 	repeat -= num;
-	memset ((char *)b, ft.pad, repeat);
+	if ((repeat == 0) && (ft.format_type == 'z'))
+	  {
+	     if (num) *(b-1) = 0;
+	  }
+	else memset ((char *)b, ft.pad, repeat);
 	SLbstring_free (bs);
 	b += repeat;
 	nitems--;
@@ -507,7 +559,7 @@ pack_according_to_format (char *format, unsigned int nitems)
    return NULL;
 }
 
-void _SLpack (void)
+void _pSLpack (void)
 {
    SLang_BString_Type *bs;
    char *fmt;
@@ -536,7 +588,25 @@ void _SLpack (void)
    SLbstring_free (bs);
 }
 
-void _SLunpack (char *format, SLang_BString_Type *bs)
+static unsigned int get_unpadded_strlen (char *str, char pad, unsigned int len)
+{
+   char *s;
+
+   s = str + len;
+   while (s > str)
+     {
+	s--;
+	if ((*s != pad) && (*s != 0))
+	  {
+	     s++;
+	     break;
+	  }
+     }
+   return (unsigned int) (s - str);
+}
+
+
+void _pSLunpack (char *format, SLang_BString_Type *bs)
 {
    Format_Type ft;
    unsigned char *b;
@@ -576,13 +646,13 @@ void _SLunpack (char *format, SLang_BString_Type *bs)
 	if (ft.is_scalar)
 	  {
 	     SLang_Array_Type *at;
-	     int dims;
+	     SLindex_Type dims;
 
 	     if (ft.repeat == 1)
 	       {
 		  SLang_Class_Type *cl;
 
-		  cl = _SLclass_get_class (ft.data_type);
+		  cl = _pSLclass_get_class (ft.data_type);
 		  memcpy ((char *)cl->cl_transfer_buf, (char *)b, ft.sizeof_type);
 		  if (ft.byteorder != NATIVE_ORDER)
 		    byteswap (ft.byteorder, (unsigned char *)cl->cl_transfer_buf, ft.sizeof_type, 1);
@@ -593,7 +663,7 @@ void _SLunpack (char *format, SLang_BString_Type *bs)
 		  continue;
 	       }
 
-	     dims = (int) ft.repeat;
+	     dims = (SLindex_Type) ft.repeat;
 	     at = SLang_create_array (ft.data_type, 0, NULL, &dims, 1);
 	     if (at == NULL)
 	       return;
@@ -609,45 +679,19 @@ void _SLunpack (char *format, SLang_BString_Type *bs)
 	     b += num_bytes;
 	     continue;
 	  }
+	
+	/* string type: s, S, or Z */
+	if (ft.format_type == 's')
+	  len = ft.repeat;
+	else
+	  len = get_unpadded_strlen ((char *)b, ft.pad, ft.repeat);
 
-	len = ft.repeat;
 	str = SLmalloc (len + 1);
 	if (str == NULL)
 	  return;
-
 	memcpy ((char *) str, (char *)b, len);
 	str [len] = 0;
 
-	if (ft.pad == ' ')
-	  {
-	     unsigned int new_len;
-
-	     s = str + len;
-	     while (s > str)
-	       {
-		  s--;
-		  if ((*s != ' ') && (*s != 0))
-		    {
-		       s++;
-		       break;
-		    }
-		  *s = 0;
-	       }
-	     new_len = (unsigned int) (s - str);
-
-	     if (new_len != len)
-	       {
-		  s = SLrealloc (str, new_len + 1);
-		  if (s == NULL)
-		    {
-		       SLfree (str);
-		       return;
-		    }
-		  str = s;
-		  len = new_len;
-	       }
-	  }
-
 	/* Avoid a bstring if possible */
 	s = SLmemchr (str, 0, len);
 	if (s == NULL)
@@ -675,7 +719,7 @@ void _SLunpack (char *format, SLang_BString_Type *bs)
      }
 }
 
-unsigned int _SLpack_compute_size (char *format)
+unsigned int _pSLpack_compute_size (char *format)
 {
    unsigned int n;
 
@@ -684,7 +728,7 @@ unsigned int _SLpack_compute_size (char *format)
    return n;
 }
 
-void _SLpack_pad_format (char *format)
+void _pSLpack_pad_format (char *format)
 {
    unsigned int len, max_len;
    Format_Type ft;
