@@ -1,9 +1,23 @@
-/* Copyright (c) 1996, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 /*--------------------------------*-C-*---------------------------------*
  * File:	slprepr.c
@@ -119,24 +133,132 @@
 /*}}}*/
 
 int (*SLprep_exists_hook) (char *, char);	/* in "slang.h" */
-int (*_SLprep_eval_hook) (char *);		/* in "_slang.h" */
+int (*_pSLprep_eval_hook) (char *);		/* in "_slang.h" */
 
-/*{{{ SLprep_open_prep (), SLprep_close_prep () */
-int SLprep_open_prep (SLPreprocess_Type *pt)
+
+struct _pSLprep_Type
 {
-   pt->this_level = 0;
-   pt->exec_level = 0;
-   pt->prev_exec_level = 0;
-   pt->comment_char = '%';
-   pt->preprocess_char = '#';
-   pt->flags = 0;
+   int this_level;
+   int exec_level;
+   int prev_exec_level;
+   char *prefix;
+   unsigned int prefix_len;
+   char *comment_start;
+   char *comment_stop;
+   unsigned int comment_start_len;
+   unsigned int flags;
+#define SLPREP_USER_FLAGS_MASK	0x00FF
+#define SLPREP_STOP_READING	0x0100
+#define SLPREP_EMBEDDED_TEXT	0x0200
+   
+   int (*exists_hook) (SLprep_Type *, char *);
+   int (*eval_hook) (SLprep_Type *, char *);
+};
+
+int SLprep_set_eval_hook (SLprep_Type *pt, int (*f)(SLprep_Type *, char *))
+{
+   if (pt == NULL)
+     return -1;
+   pt->eval_hook = f;
+   return 0;
+}
+int SLprep_set_exists_hook (SLprep_Type *pt, int (*f)(SLprep_Type *, char *))
+{
+   if (pt == NULL)
+     return -1;
+   pt->exists_hook = f;
    return 0;
 }
 
-void SLprep_close_prep (SLPreprocess_Type *pt)
+void SLprep_delete (SLprep_Type *pt)
 {
-   (void) pt;
+   if (pt == NULL)
+     return;
+   
+   /* NULLs ok */
+   SLang_free_slstring (pt->comment_start);
+   SLang_free_slstring (pt->comment_stop);
+   SLang_free_slstring (pt->prefix);
+   
+   SLfree ((char *) pt);
 }
+
+int SLprep_set_comment (SLprep_Type *pt, char *start, char *stop)
+{
+   if ((pt == NULL) || (start == NULL))
+     return -1;
+
+   if (NULL == (start = SLang_create_slstring (start)))
+     return -1;
+   
+   if (stop == NULL)
+     stop = "";
+
+   if (NULL == (stop = SLang_create_slstring (stop)))
+     {
+	SLang_free_slstring (start);
+	return -1;
+     }
+
+   if (pt->comment_start != NULL)
+     SLang_free_slstring (pt->comment_start);
+   pt->comment_start = start;
+   pt->comment_start_len = strlen (start);
+
+   if (pt->comment_stop != NULL)
+     SLang_free_slstring (pt->comment_stop); 
+   pt->comment_stop = stop;
+   
+   return 0;
+}
+
+int SLprep_set_prefix (SLprep_Type *pt, char *prefix)
+{
+   if ((pt == NULL) || (prefix == NULL))
+     return -1;
+   
+   if (NULL == (prefix = SLang_create_slstring (prefix)))
+     return -1;
+   
+   if (pt->prefix != NULL)
+     SLang_free_slstring (pt->prefix);
+   pt->prefix = prefix;
+   pt->prefix_len = strlen (prefix);
+   return 0;
+}
+
+/*{{{ SLprep_open_prep (), SLprep_close_prep () */
+SLprep_Type *SLprep_new (void)
+{
+   SLprep_Type *pt;
+   
+   if (NULL == (pt = (SLprep_Type *)SLcalloc (1, sizeof (SLprep_Type))))
+     return NULL;
+   
+   if (-1 == SLprep_set_comment (pt, "%", ""))
+     {
+	SLprep_delete (pt);
+	return NULL;
+     }
+   if (-1 == SLprep_set_prefix (pt, "#"))
+     {
+	SLprep_delete (pt);
+	return NULL;
+     }
+   return pt;
+}
+
+int SLprep_set_flags (SLprep_Type *pt, unsigned int flags)
+{
+   if (pt == NULL)
+     return -1;
+
+   flags &= SLPREP_USER_FLAGS_MASK;
+   pt->flags &= ~SLPREP_USER_FLAGS_MASK;
+   pt->flags |= flags;
+   return 0;
+}
+
 /*}}}*/
 
 /*{{{ SLwildcard () */
@@ -202,7 +324,7 @@ static int SLwildcard (char *pattern, char *string)
 #endif
 
 /* The extra one is for NULL termination */
-char *_SLdefines [MAX_DEFINES + 1];
+char *_pSLdefines [MAX_DEFINES + 1];
 
 int SLdefine_for_ifdef (char *s)	/*{{{*/
 {
@@ -210,7 +332,7 @@ int SLdefine_for_ifdef (char *s)	/*{{{*/
 
    for (i = 0; i < MAX_DEFINES; i++)
      {
-	char *s1 = _SLdefines [i];
+	char *s1 = _pSLdefines [i];
 
 	if (s1 == s)
 	  return 0;		       /* already defined (hashed string) */
@@ -222,7 +344,7 @@ int SLdefine_for_ifdef (char *s)	/*{{{*/
 	if (s == NULL)
 	  return -1;
 
-	_SLdefines[i] = s;
+	_pSLdefines[i] = s;
 	return 0;
      }
    return -1;
@@ -230,11 +352,14 @@ int SLdefine_for_ifdef (char *s)	/*{{{*/
 /*}}}*/
 
 /*{{{ static functions */
-static int is_any_defined(char *buf, char comment)	/*{{{*/
+static int is_any_defined (SLprep_Type *pt, char *buf)	/*{{{*/
 {
    char *sys;
    unsigned int i;
-
+   char comment;
+   
+   /* FIXME: priority: low.  This needs adapted to multi-char comments */
+   comment = pt->comment_start[0];
    while (1)
      {
 	register char ch;
@@ -245,9 +370,9 @@ static int is_any_defined(char *buf, char comment)	/*{{{*/
 
 	if ((ch == '\n') || (ch == 0) || (ch == comment))
 	  return 0;
-
+	
 	i = 0;
-	while (NULL != (sys = _SLdefines [i++]))
+	while (NULL != (sys = _pSLdefines [i++]))
 	  {
 	     unsigned int n;
 
@@ -259,9 +384,11 @@ static int is_any_defined(char *buf, char comment)	/*{{{*/
 	       {
 		  char ch1 = *(buf + n);
 
-		  if ((ch1 == '\n') || (ch1 == 0) ||
-		      (ch1 == ' ') || (ch1 == '\t') || (ch1 == comment))
+		  if ((ch1 == '\n') || (ch1 == 0)
+		      || (ch1 == ' ') || (ch1 == '\t')
+		      || (ch == comment))
 		    return 1;
+		  
 	       }
 	  }
 
@@ -294,11 +421,15 @@ static unsigned char *tokenize (unsigned char *buf, char *token, unsigned int le
    return buf;
 }
 
-static int is_env_defined (char *buf, char comment)	/*{{{*/
+static int is_env_defined (SLprep_Type *pt, char *buf)	/*{{{*/
 {
-   char * env, token [32];
+   char *env, token [128];
+   char comment = pt->comment_start[0];
 
-   if ((*buf <= ' ') || (*buf == comment)) return 0;	/* no token */
+   /* FIXME: priority: low.  This needs adapted to multi-char comments */
+   if (((unsigned char)*buf <= ' ') 
+       || (*buf == comment)) 
+     return 0;	/* no token */
 
    if (NULL == (buf = (char *) tokenize ((unsigned char *) buf,
 					 token, sizeof (token))))
@@ -325,7 +456,7 @@ static int is_env_defined (char *buf, char comment)	/*{{{*/
 /*}}}*/
 /*}}}*/
 
-int SLprep_line_ok (char *buf, SLPreprocess_Type *pt)	/*{{{*/
+int SLprep_line_ok (char *buf, SLprep_Type *pt)	/*{{{*/
 {
    int level, prev_exec_level, exec_level;
    unsigned int flags;
@@ -342,27 +473,32 @@ int SLprep_line_ok (char *buf, SLPreprocess_Type *pt)	/*{{{*/
    prev_exec_level = pt->prev_exec_level;
    flags = pt->flags;
 
-   if (*buf != pt->preprocess_char)
+   if ((*buf != pt->prefix[0])
+       || (0 != strncmp (buf, pt->prefix, pt->prefix_len)))
      {
 	/* ignore out-of-context or embedded text */
 	if ((level != exec_level) || (flags & SLPREP_EMBEDDED_TEXT))
 	  return 0;
 
-	if (*buf == '\n') return flags & SLPREP_BLANK_LINES_OK;
-	if (*buf == pt->comment_char) return flags & SLPREP_COMMENT_LINES_OK;
-	
+	if (*buf == '\n') 
+	  return (0 != (flags & SLPREP_BLANK_LINES_OK));
+
+	if ((*buf == pt->comment_start[0])
+	    && (0 == strncmp (buf, pt->comment_start, pt->comment_start_len)))
+	  return (0 != (flags & SLPREP_COMMENT_LINES_OK));
+
 	return 1;
      }
    
-   buf++;
-   
+   buf += pt->prefix_len;
+
    /*
     * Always allow '#!' to pass. This could be a shell script
     * with something like '#! /usr/local/bin/slang'
     */
-   if ((*buf == '!') && (pt->preprocess_char == '#'))
+   if ((*buf == '!') && (pt->prefix[0] == '#') && (pt->prefix_len == 1))
      return 0;
-   
+
    /* Allow whitespace as in '#  ifdef'  */
    while ((*buf == ' ') || (*buf == '\t')) buf++;
    
@@ -460,7 +596,7 @@ int SLprep_line_ok (char *buf, SLPreprocess_Type *pt)	/*{{{*/
 	/* for 'ifdef' and 'ifndef' we are done */
 	if (!strncmp (buf, "def", 3))
 	  {
-	     test = is_any_defined(buf + 3, pt->comment_char);
+	     test = is_any_defined(pt, buf + 3);
 	  } 
 	else 
 	  {
@@ -484,16 +620,16 @@ int SLprep_line_ok (char *buf, SLPreprocess_Type *pt)	/*{{{*/
 	       }
 	     
 	     if (*buf == '$') 
-	       test = is_env_defined (buf + 1, pt->comment_char);
-	     else if (*buf == '('
-		      && (_SLprep_eval_hook != NULL)) 
-	       test = (*_SLprep_eval_hook) (buf);
+	       test = is_env_defined (pt, buf + 1);
+	     else if ((*buf == '(')
+		      && (pt->eval_hook != NULL))
+	       test = pt->eval_hook (pt, buf);
 	     else if (!strncmp (buf, "eval", 4)
-		      && (_SLprep_eval_hook != NULL))
-	       test = (*_SLprep_eval_hook) (buf + 4);
+		      && (pt->eval_hook != NULL))
+	       test = pt->eval_hook (pt, buf + 4);
 	     else if (!strncmp (buf, "exists", 6)
-		      && (SLprep_exists_hook != NULL))
-	       test = (*SLprep_exists_hook) (buf + 6, pt->comment_char);
+		      && (pt->exists_hook != NULL))
+	       test = pt->exists_hook (pt, buf + 6);
 	     else if (!strncmp (buf, "true", 4))
 	       test = 1;
 	     else if (strncmp (buf, "false", 5))
@@ -524,7 +660,7 @@ int main (int argc, char *argv[])
 {
    int i;
    char buf[1024];
-   SLPreprocess_Type pt;
+   SLprep_Type pt;
 
    SLprep_open_prep (&pt);
    SLdefine_for_ifdef ("UNIX");
