@@ -1,15 +1,53 @@
 /* ed style regular expressions */
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
 #include "slang.h"
 #include "_slang.h"
 
+struct _pSLRegexp_Type
+{
+   /* These must be set by calling routine. */
+   unsigned char *pat;		       /* regular expression pattern */
+   unsigned char *buf;		       /* buffer for compiled regexp */
+   unsigned int buf_len;	       /* length of buffer */
+   int case_sensitive;		       /* 1 if match is case sensitive  */
+
+   /* The rest are set by SLang_regexp_compile */
+
+   int must_match;		       /* 1 if line must contain substring */
+   int must_match_bol;		       /* true if it must match beginning of line */
+   unsigned char must_match_str[16];   /* 15 char null term substring */
+   int osearch;			       /* 1 if ordinary search suffices */
+   unsigned int min_length;	       /* minimum length the match must be */
+   int beg_matches[10];		       /* offset of start of \( */
+   unsigned int end_matches[10];       /* length of nth submatch
+					* Note that the entire match corresponds
+					* to \0
+					*/
+   int offset;			       /* offset to be added to beg_matches */
+};
+
 #define SET_BIT(b, n) b[(unsigned int) (n) >> 3] |= 1 << ((unsigned int) (n) % 8)
 #define TEST_BIT(b, n) (b[(unsigned int)(n) >> 3] & (1 << ((unsigned int) (n) % 8)))
 #define LITERAL 1
@@ -77,7 +115,10 @@ static unsigned char *do_nth_match (Re_Context_Type *ctx, int n, unsigned char *
 }
 
 /* returns pointer to the end of regexp or NULL */
-static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned char *str, unsigned char *estr, unsigned char *buf, register int cs)
+static unsigned char *regexp_looking_at (Re_Context_Type *ctx, 
+					 unsigned char *str, unsigned char *estr, 
+					 unsigned char *regexp, 
+					 int cs)
 {
    register unsigned char p, p1;
    unsigned char *save_str, *tmpstr;
@@ -85,11 +126,11 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
    int save_num_open;
    char save_closed_matches[10];
 
-   p = *buf++;
+   p = *regexp++;
 
    while (p != 0)
      {
-	/* p1 = UPPERCASE(*buf); */
+	/* p1 = UPPERCASE(*regexp); */
 	/* if (str < estr) c = UPPERCASE(*str); */
 
 	switch((unsigned char) p)
@@ -129,33 +170,33 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 	     break;
 #ifdef NOT_LITERAL
 	   case NOT_LITERAL:
-	     if ((str >= estr) || (*buf == UPPERCASE(*str))) return (NULL);
-	     str++; buf++;
+	     if ((str >= estr) || (*regexp == UPPERCASE(*str))) return (NULL);
+	     str++; regexp++;
 	     break;
 
 	   case MAYBE_ONCE | NOT_LITERAL:
 	     save_str = str;
-	     if ((str < estr) && (*buf != UPPERCASE(*str))) str++;
-	     buf++;
+	     if ((str < estr) && (*regexp != UPPERCASE(*str))) str++;
+	     regexp++;
 	     goto match_rest;
 
 	   case NOT_LITERAL | LEAST_ONCE:   /* match at least once */
-	     if ((str >= estr) || (UPPERCASE(*str) == UPPERCASE(*buf))) return (NULL);
+	     if ((str >= estr) || (UPPERCASE(*str) == UPPERCASE(*regexp))) return (NULL);
 	     str++;
 	     /* drop */
 	   case STAR | NOT_LITERAL:
-	     save_str = str;  p1 = *buf;
+	     save_str = str;  p1 = *regexp;
 	     while ((str < estr) && (UPPERCASE(*str) != p1)) str++;
-	     buf++;
+	     regexp++;
 	     goto match_rest;
 
 	     /* this type consists of the expression + two bytes that
 	        determine number of matches to perform */
 	   case MANY | NOT_LITERAL:
-	     p1 = *buf; buf++;
-	     n = n0 = (int) (unsigned char) *buf++;
+	     p1 = *regexp; regexp++;
+	     n = n0 = (int) (unsigned char) *regexp++;
 	     /* minimum number to match--- could be 0 */
-	     n1 = (int) (unsigned char) *buf++;
+	     n1 = (int) (unsigned char) *regexp++;
 	     /* maximum number to match */
 
 	     while (n && (str < estr) && (p1 != UPPERCASE(*str)))
@@ -175,33 +216,33 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 	     goto match_rest;
 #endif				       /* NOT_LITERAL */
 	   case LITERAL:
-	     if ((str >= estr) || (*buf != UPPERCASE(*str))) return (NULL);
-	     str++; buf++;
+	     if ((str >= estr) || (*regexp != UPPERCASE(*str))) return (NULL);
+	     str++; regexp++;
 	     break;
 
 	   case MAYBE_ONCE | LITERAL:
 	     save_str = str;
-	     if ((str < estr) && (*buf == UPPERCASE(*str))) str++;
-	     buf++;
+	     if ((str < estr) && (*regexp == UPPERCASE(*str))) str++;
+	     regexp++;
 	     goto match_rest;
 
 	   case LITERAL | LEAST_ONCE:   /* match at least once */
-	     if ((str >= estr) || (UPPERCASE(*str) != UPPERCASE(*buf))) return (NULL);
+	     if ((str >= estr) || (UPPERCASE(*str) != UPPERCASE(*regexp))) return (NULL);
 	     str++;
 	     /* drop */
 	   case STAR | LITERAL:
-	     save_str = str;  p1 = *buf;
+	     save_str = str;  p1 = *regexp;
 	     while ((str < estr) && (UPPERCASE(*str) == p1)) str++;
-	     buf++;
+	     regexp++;
 	     goto match_rest;
 
 	     /* this type consists of the expression + two bytes that
 	        determine number of matches to perform */
 	   case MANY | LITERAL:
-	     p1 = *buf; buf++;
-	     n = n0 = (int) (unsigned char) *buf++;
+	     p1 = *regexp; regexp++;
+	     n = n0 = (int) (unsigned char) *regexp++;
 	     /* minimum number to match--- could be 0 */
-	     n1 = (int) (unsigned char) *buf++;
+	     n1 = (int) (unsigned char) *regexp++;
 	     /* maximum number to match */
 
 	     while (n && (str < estr) && (p1 == UPPERCASE(*str)))
@@ -221,14 +262,14 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 	     goto match_rest;
 
 	   case NTH_MATCH:
-	     if ((str = do_nth_match(ctx, (int) (unsigned char) *buf, str, estr)) == NULL) return(NULL);
-	     buf++;
+	     if ((str = do_nth_match(ctx, (int) (unsigned char) *regexp, str, estr)) == NULL) return(NULL);
+	     regexp++;
 	     break;
 
 	   case MAYBE_ONCE | NTH_MATCH:
 	     save_str = str;
-	     tmpstr = do_nth_match (ctx, (int) (unsigned char) *buf, str, estr);
-	     buf++;
+	     tmpstr = do_nth_match (ctx, (int) (unsigned char) *regexp, str, estr);
+	     regexp++;
 	     if (tmpstr != NULL)
 	       {
 		  str = tmpstr;
@@ -237,15 +278,15 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 	     continue;
 
 	   case LEAST_ONCE | NTH_MATCH:
-	     if ((str = do_nth_match(ctx, (int) (unsigned char) *buf, str, estr)) == NULL) return(NULL);
+	     if ((str = do_nth_match(ctx, (int) (unsigned char) *regexp, str, estr)) == NULL) return(NULL);
 	     /* drop */
 	   case STAR | NTH_MATCH:
 	     save_str = str;
-	     while (NULL != (tmpstr = do_nth_match(ctx, (int) (unsigned char) *buf, str, estr)))
+	     while (NULL != (tmpstr = do_nth_match(ctx, (int) (unsigned char) *regexp, str, estr)))
 	       {
 		  str = tmpstr;
 	       }
-	     buf++;
+	     regexp++;
 	     goto match_rest;
 
 	   case MANY | NTH_MATCH: return(NULL);
@@ -253,24 +294,24 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 
 	   case RANGE:
 	     if (str >= estr) return (NULL);
-	     if (TEST_BIT(buf, UPPERCASE(*str)) == 0) return (NULL);
-	     buf += 32; str++;
+	     if (TEST_BIT(regexp, UPPERCASE(*str)) == 0) return (NULL);
+	     regexp += 32; str++;
 	     break;
 
 	   case MAYBE_ONCE | RANGE:
 	     save_str = str;
-	     if ((str < estr) && TEST_BIT(buf, UPPERCASE(*str))) str++;
-	     buf += 32;
+	     if ((str < estr) && TEST_BIT(regexp, UPPERCASE(*str))) str++;
+	     regexp += 32;
 	     goto match_rest;
 
 	   case LEAST_ONCE | RANGE:
-	     if ((str >= estr) || (0 == TEST_BIT(buf, UPPERCASE(*str)))) return NULL;
+	     if ((str >= estr) || (0 == TEST_BIT(regexp, UPPERCASE(*str)))) return NULL;
 	     str++;
 	     /* drop */
 	   case STAR | RANGE:
 	     save_str = str;
-	     while ((str < estr) && TEST_BIT(buf, UPPERCASE(*str))) str++;
-	     buf += 32;
+	     while ((str < estr) && TEST_BIT(regexp, UPPERCASE(*str))) str++;
+	     regexp += 32;
 	     goto match_rest;
 
 	     /* The first 32 bytes correspond to the range and the two
@@ -278,11 +319,11 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 	      */
 	   case MANY | RANGE:
 	     /* minimum number to match--- could be 0 */
-	     n = n0 = (int) (unsigned char) *(buf + 32);
+	     n = n0 = (int) (unsigned char) *(regexp + 32);
 	     /* maximum number to match */
-	     n1 = (int) (unsigned char) *(buf + 33);
+	     n1 = (int) (unsigned char) *(regexp + 33);
 
-	     while (n && (str < estr) && (TEST_BIT(buf, UPPERCASE(*str))))
+	     while (n && (str < estr) && (TEST_BIT(regexp, UPPERCASE(*str))))
 	       {
 		  n--;
 		  str++;
@@ -290,12 +331,12 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 	     if (n) return (NULL);
 	     save_str = str;
 	     n = n1 - n0;
-	     while (n && (str < estr) && (TEST_BIT(buf, UPPERCASE(*str))))
+	     while (n && (str < estr) && (TEST_BIT(regexp, UPPERCASE(*str))))
 	       {
 		  n--;
 		  str++;
 	       }
-	     buf += 34;		       /* 32 + 2 */
+	     regexp += 34;		       /* 32 + 2 */
 	     goto match_rest;
 
 	   case ANY_DIGIT:
@@ -350,13 +391,13 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 
 	   default: return (NULL);
 	  }
-	p = *buf++;
+	p = *regexp++;
 	continue;
 
 	match_rest:
 	if (save_str == str)
 	  {
-	     p = *buf++;
+	     p = *regexp++;
 	     continue;
 	  }
 
@@ -370,7 +411,7 @@ static unsigned char *regexp_looking_at (Re_Context_Type *ctx, register unsigned
 	save_num_open = ctx->open_paren_number;
 	while (str >= save_str)
 	  {
-	     tmpstr = regexp_looking_at (ctx, str, estr, buf, cs);
+	     tmpstr = regexp_looking_at (ctx, str, estr, regexp, cs);
 	     if (tmpstr != NULL) return(tmpstr);
 	     SLMEMCPY(ctx->closed_paren_matches, save_closed_matches, sizeof(ctx->closed_paren_matches));
 	     ctx->open_paren_number = save_num_open;
@@ -417,7 +458,7 @@ static void init_re_context (Re_Context_Type *ctx, SLRegexp_Type *reg,
    ctx->len = len;
 }
 
-unsigned char *SLang_regexp_match(unsigned char *str,
+static unsigned char *regexp_match(unsigned char *str,
 				  unsigned int len, SLRegexp_Type *reg)
 {
    register unsigned char c = 0, *estr = str + len;
@@ -484,6 +525,11 @@ unsigned char *SLang_regexp_match(unsigned char *str,
    return NULL;
 }
 
+char *SLregexp_match (SLRegexp_Type *reg, char *str, unsigned int len)
+{
+   return (char *) regexp_match ((unsigned char *)str, len, reg);
+}
+
 static unsigned char *convert_digit(unsigned char *pat, int *nn)
 {
    int n = 0, m = 0;
@@ -505,7 +551,7 @@ static unsigned char *convert_digit(unsigned char *pat, int *nn)
 #define ERROR  return (int) (pat - reg->pat)
 
 /* Returns 0 if successful or offset in pattern of error */
-int SLang_regexp_compile (SLRegexp_Type *reg)
+static int regexp_compile (SLRegexp_Type *reg)
 {
    register unsigned char *buf, *ebuf, *pat;
    unsigned char *last = NULL, *tmppat;
@@ -579,7 +625,7 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
      {
 	if (buf >= ebuf - 3)
 	  {
-	     SLang_doerror ("Pattern too large to be compiled.");
+	     SLang_verror (SL_BUILTIN_LIMIT_EXCEEDED, "Pattern too large to be compiled.");
 	     ERROR;
 	  }
 
@@ -857,6 +903,41 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
    return(0);
 }
 
+void SLregexp_free (SLRegexp_Type *reg)
+{
+   if (reg == NULL)
+     return;
+   if (reg->buf != NULL)
+     SLfree ((char *) reg->buf);
+   SLfree ((char *) reg);
+}
+
+SLRegexp_Type *SLregexp_compile (char *pattern, unsigned int flags)
+{
+   SLRegexp_Type *reg;
+
+   reg = (SLRegexp_Type *)SLcalloc (1, sizeof (SLRegexp_Type));
+   if (reg == NULL)
+     return NULL;
+   
+   if (NULL == (reg->buf = (unsigned char *)SLmalloc (1024)))
+     {
+	SLfree ((char *) reg);
+	return NULL;
+     }
+   reg->buf_len = 1024;
+   reg->case_sensitive = (0 == (flags & SLREGEXP_CASELESS));
+   reg->pat = (unsigned char *)pattern;
+
+   if (-1 == regexp_compile (reg))
+     {
+	SLregexp_free (reg);
+	return NULL;
+     }
+   
+   return reg;
+}
+
 char *SLregexp_quote_string (char *re, char *buf, unsigned int buflen)
 {
    char ch;
@@ -895,6 +976,39 @@ char *SLregexp_quote_string (char *re, char *buf, unsigned int buflen)
    return NULL;
 }
 
+int SLregexp_nth_match (SLRegexp_Type *reg, unsigned int nth,
+			unsigned int *ofsp, unsigned int *lenp)
+{
+   if (nth >= 10)
+     {
+	SLang_set_error (SL_InvalidParm_Error);
+	return -1;
+     }
+   if (reg->beg_matches[nth] < 0)
+     return -1;
+   
+   if (ofsp != NULL)
+     *ofsp = (unsigned int) reg->beg_matches[nth];
+   if (lenp != NULL)
+     *lenp = (unsigned int) reg->end_matches[nth];
+
+   return 0;
+}
+
+int SLregexp_get_hints (SLRegexp_Type *reg, unsigned int *hintsp)
+{
+   unsigned int hints = 0;
+   
+   if (reg == NULL)
+     return -1;
+   
+   if (reg->osearch) hints |= SLREGEXP_HINT_OSEARCH;
+   if (reg->must_match_bol) hints |= SLREGEXP_HINT_BOL;
+   
+   *hintsp = hints;
+   return 0;
+}
+
 #if 0
 #define MAX_EXP 4096
 int main(int argc, char **argv)
