@@ -27,12 +27,13 @@ USA.
 #define MAX_TOKEN_LEN 254
 #define MAX_FILE_LINE_LEN 256
 
-int _pSLang_Compile_Line_Num_Info;
+/* int _pSLang_Compile_Line_Num_Info; */
 #if SLANG_HAS_BOSEOS
 int _pSLang_Compile_BOSEOS;
+int _pSLang_Compile_BOFEOF;
 #endif
 #if SLANG_HAS_DEBUG_CODE
-static int Default_Compile_Line_Num_Info;
+/* static int Default_Compile_Line_Num_Info; */
 #if 0
 static int Default_Compile_BOSEOS;
 #endif
@@ -97,8 +98,27 @@ static char *map_token_to_string (_pSLang_Token_Type *tok)
       case CPAREN_TOKEN: s = ")"; break;
       case OBRACE_TOKEN: s = "{"; break;
       case CBRACE_TOKEN: s = "}"; break;
-      case DEREF_TOKEN: s = "@"; break;
+      case POW_TOKEN: s = "^"; break;
+      case ADD_TOKEN: s = "+"; break;
+      case SUB_TOKEN: s = "-"; break;
+      case TIMES_TOKEN: s = "*"; break;
+      case DIV_TOKEN: s = "/"; break;
+      case LT_TOKEN: s = "<"; break;
+      case LE_TOKEN: s = "<="; break;
+      case GT_TOKEN: s = ">"; break;
+      case GE_TOKEN: s = ">="; break;
+      case EQ_TOKEN: s = "=="; break;
+      case NE_TOKEN: s = "!="; break;
+      case AND_TOKEN: s = "and"; break;
+      case OR_TOKEN: s = "or"; break;
+      case MOD_TOKEN: s = "mod"; break;
+      case BAND_TOKEN: s = "&"; break;
+      case SHL_TOKEN: s = "shl"; break;
+      case SHR_TOKEN: s = "shr"; break;
+      case BXOR_TOKEN: s = "xor"; break;
+      case BOR_TOKEN: s = "|"; break;
       case POUND_TOKEN: s = "#"; break;
+      case DEREF_TOKEN: s = "@"; break;
       case COMMA_TOKEN: s = ","; break;
       case SEMICOLON_TOKEN: s = ";"; break;
       case COLON_TOKEN: s = ":"; break;
@@ -135,7 +155,8 @@ void _pSLparse_error (int errcode, char *str, _pSLang_Token_Type *tok, int flag)
      str = "Parse Error";
 
 #if SLANG_HAS_DEBUG_CODE
-   if (tok != NULL) line = tok->line_number;
+   if ((tok != NULL) && (tok->line_number != -1))
+     line = tok->line_number;
 #endif
    if (file == NULL) file = "??";
 
@@ -1056,10 +1077,11 @@ int SLang_load_object (SLang_Load_Type *x)
    SLang_Load_Type *save_llt;
    char *save_input_line, *save_input_line_ptr;
 #if SLANG_HAS_DEBUG_CODE
-   int save_compile_line_num_info;
+   /* int save_compile_line_num_info; */
 #endif
 #if SLANG_HAS_BOSEOS
    int save_compile_boseos;
+   int save_compile_bofeof;
 #endif
    int save_auto_declare_variables;
 
@@ -1075,10 +1097,11 @@ int SLang_load_object (SLang_Load_Type *x)
      }
 
 #if SLANG_HAS_DEBUG_CODE
-   save_compile_line_num_info = _pSLang_Compile_Line_Num_Info;
+   /* save_compile_line_num_info = _pSLang_Compile_Line_Num_Info; */
 #endif
 #if SLANG_HAS_BOSEOS
    save_compile_boseos = _pSLang_Compile_BOSEOS;
+   save_compile_bofeof = _pSLang_Compile_BOFEOF;
 #endif
    save_this_pp = This_SLpp;
    save_input_line = Input_Line;
@@ -1095,7 +1118,7 @@ int SLang_load_object (SLang_Load_Type *x)
    _pSLang_Auto_Declare_Globals = x->auto_declare_globals;
 
 #if SLANG_HAS_DEBUG_CODE
-   _pSLang_Compile_Line_Num_Info = Default_Compile_Line_Num_Info;
+   /* _pSLang_Compile_Line_Num_Info = Default_Compile_Line_Num_Info; */
 #endif
 #if SLANG_HAS_BOSEOS
 #if 0
@@ -1138,10 +1161,11 @@ int SLang_load_object (SLang_Load_Type *x)
    This_SLpp = save_this_pp;
 
 #if SLANG_HAS_DEBUG_CODE
-   _pSLang_Compile_Line_Num_Info = save_compile_line_num_info;
+   /* _pSLang_Compile_Line_Num_Info = save_compile_line_num_info; */
 #endif
 #if SLANG_HAS_BOSEOS
    _pSLang_Compile_BOSEOS = save_compile_boseos;
+   _pSLang_Compile_BOFEOF = save_compile_bofeof;
 #endif
    if (_pSLang_Error) return -1;
    return 0;
@@ -1426,10 +1450,13 @@ void _pSLcompile_byte_compiled (void)
 	       return;
 	     tok.v.long_val = atol (buf);
 	     break;
-#ifdef X___HAVE_LONG_LONG
+#ifdef HAVE_LONG_LONG
 	   case LLONG_TOKEN:
 	   case ULLONG_TOKEN:
-	     XXX
+	     if (NULL == check_byte_compiled_token (buf))
+	       return;
+	     tok.v.llong_val = SLatoll ((unsigned char *)buf);
+	     break;
 #endif
 	   case COMPLEX_TOKEN:
 	   case FLOAT_TOKEN:
@@ -1439,6 +1466,15 @@ void _pSLcompile_byte_compiled (void)
 	     tok.v.s_val = buf;
 	     break;
 
+	   case ESC_STRING_DOLLAR_TOKEN:
+	     if (NULL == (ebuf = check_byte_compiled_token (buf)))
+	       return;
+	     tok.v.s_val = buf;
+	     (void) expand_escaped_string (buf, buf, ebuf, &len, 0);
+	     tok.hash = _pSLstring_hash ((unsigned char *)buf, (unsigned char *)buf + len);
+	     type = STRING_DOLLAR_TOKEN;
+	     break;
+
 	   case ESC_STRING_TOKEN:
 	     if (NULL == (ebuf = check_byte_compiled_token (buf)))
 	       return;
@@ -1463,6 +1499,8 @@ void _pSLcompile_byte_compiled (void)
 	   case DEFINE_PRIVATE_TOKEN:
 	   case DEFINE_PUBLIC_TOKEN:
 	   case DOT_TOKEN:
+	   case DOT_METHOD_CALL_TOKEN:
+	   case STRING_DOLLAR_TOKEN:
 	   case STRING_TOKEN:
 	   case IDENT_TOKEN:
 	   case _REF_TOKEN:
@@ -1623,6 +1661,15 @@ static void byte_compile_token (_pSLang_Token_Type *tok)
 	sprintf (b3, "%lu", tok->v.long_val);
 	break;
 
+#ifdef HAVE_LONG_LONG
+      case LLONG_TOKEN:
+	sprintf (b3, "%lld", tok->v.llong_val);
+	break;
+	
+      case ULLONG_TOKEN:
+	sprintf (b3, "%llu", tok->v.ullong_val);
+	break;
+#endif
       case _BSTRING_TOKEN:
 	s = (unsigned char *) tok->v.s_val;
 	len = (unsigned int) tok->hash;
@@ -1646,6 +1693,7 @@ static void byte_compile_token (_pSLang_Token_Type *tok)
 	buf[0] = ESC_BSTRING_TOKEN;
 	break;
 
+      case STRING_DOLLAR_TOKEN:
       case STRING_TOKEN:
 	s = (unsigned char *)tok->v.s_val;
 
@@ -1655,7 +1703,8 @@ static void byte_compile_token (_pSLang_Token_Type *tok)
 	    return;
 
 	if (is_escaped)
-	  buf[0] = ESC_STRING_TOKEN;
+	  buf[0] = ((tok->type == STRING_TOKEN) 
+		    ? ESC_STRING_TOKEN : ESC_STRING_DOLLAR_TOKEN);
 	break;
 
       /* case _DEREF_ASSIGN_TOKEN: */
@@ -1672,6 +1721,7 @@ static void byte_compile_token (_pSLang_Token_Type *tok)
       case _SCALAR_MINUSMINUS_TOKEN:
       case _SCALAR_POST_MINUSMINUS_TOKEN:
       case DOT_TOKEN:
+      case DOT_METHOD_CALL_TOKEN:
       case TMP_TOKEN:
       case DEFINE_TOKEN:
       case DEFINE_STATIC_TOKEN:
@@ -1753,8 +1803,10 @@ int SLang_byte_compile_file (char *name, int method)
 int SLang_generate_debug_info (int x)
 {
 #if SLANG_HAS_DEBUG_CODE
-   int y = Default_Compile_Line_Num_Info;
-   Default_Compile_Line_Num_Info = x;
+   /* int y = Default_Compile_Line_Num_Info; */
+   /* Default_Compile_Line_Num_Info = x; */
+   int y = 0;
+   (void)x;
 #if 0
    if (x == 0)
      Default_Compile_BOSEOS = 0;
