@@ -1,4 +1,6 @@
 #include <stdio.h>
+#include <string.h>
+#include <stdlib.h>
 #include <slang.h>
 #include <math.h>
 
@@ -105,10 +107,21 @@ static void get_c_struct_via_ref (SLang_Ref_Type *r)
    (void) SLang_assign_cstruct_to_ref (r, (VOID_STAR) &C_Struct_Buf, C_Struct);
 }
 
-
+static void test_pop_mmt (void)
+{
+   SLang_MMT_Type *mmt;
    
+   if (NULL == (mmt = SLang_pop_mmt (SLang_peek_at_stack ())))
+     return;
+   
+   if (-1 == SLang_push_mmt (mmt))
+     SLang_free_mmt (mmt);
+}
+
+static void fake_import (char *);
 static SLang_Intrin_Fun_Type Intrinsics [] =
 {
+   MAKE_INTRINSIC_S("fake_import", fake_import, VOID_TYPE),
    MAKE_INTRINSIC_I("exit", c_exit, VOID_TYPE),
    MAKE_INTRINSIC_1("test_char_return", test_char_return, SLANG_CHAR_TYPE, SLANG_CHAR_TYPE),
    MAKE_INTRINSIC_1("test_short_return", test_short_return, SLANG_SHORT_TYPE, SLANG_SHORT_TYPE),
@@ -118,23 +131,48 @@ static SLang_Intrin_Fun_Type Intrinsics [] =
 #if SLANG_HAS_FLOAT
    MAKE_INTRINSIC_1("test_double_return", test_double_return, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE),
 #endif
+   MAKE_INTRINSIC_0("test_pop_mmt", test_pop_mmt, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("get_c_struct", get_c_struct, VOID_TYPE),
    MAKE_INTRINSIC_0("set_c_struct", set_c_struct, VOID_TYPE),
    MAKE_INTRINSIC_1("get_c_struct_via_ref", get_c_struct_via_ref, VOID_TYPE, SLANG_REF_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
 
+static void fake_import (char *ns_name)
+{
+   SLang_NameSpace_Type *ns = SLns_create_namespace (ns_name);
+   if (ns == NULL)
+     return;
+   
+   (void) SLns_add_intrin_fun_table (ns, Intrinsics, NULL);
+}
 
 int main (int argc, char **argv)
 {
    int i;
+   int utf8 = 0;
 
-   if (argc < 2)
+   for (i = 1; i < argc; i++)
      {
-	fprintf (stderr, "Usage: %s FILE...\n", argv[0]);
+	char *arg = argv[i];
+	if (*arg != '-')
+	  break;
+
+	if (0 == strcmp (arg, "-utf8"))
+	  {
+	     utf8 = 1;
+	     continue;
+	  }
+	i = argc;
+     }
+
+   if (i >= argc)
+     {
+	fprintf (stderr, "Usage: %s [-utf8] FILE...\n", argv[0]);
 	return 1;
      }
-   
+   (void) SLutf8_enable (utf8);
+
    if ((-1 == SLang_init_all ())
        || (-1 == SLang_init_array_extra ())
        || (-1 == SLadd_intrin_fun_table (Intrinsics, NULL)))
@@ -167,16 +205,29 @@ int main (int argc, char **argv)
    fpsetmask (~(FP_X_OFL|FP_X_INV|FP_X_DZ|FP_X_DNML|FP_X_UFL|FP_X_IMP));
 #endif
 
-   if (argc > 2)
+   if (i + 1 < argc)
      Ignore_Exit = 1;
 
-   for (i = 1; i < argc; i++)
+   while (i < argc)
      {
-	if (-1 == SLang_load_file (argv[i]))
+	char *file = argv[i];
+	if (0 == strncmp (SLpath_extname (file), ".slc", 4))
+	  {
+	     char *file_sl = SLmake_string (file);
+	     file_sl[strlen(file_sl)-1] = 0;
+	     if (-1 == SLang_byte_compile_file (file_sl, 0))
+	       {
+		  SLfree (file_sl);
+		  return 1;
+	       }
+	     SLfree (file_sl);
+	  }
+	if (-1 == SLang_load_file (file))
 	  return 1;
+	i++;
      }
 
-   return SLang_Error;
+   return SLang_get_error ();
 }
 
 	
